/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_DBMS_H
#define _UDM_DBMS_H


typedef enum
{
  UDM_DBCMD_CREATE=               0,
  UDM_DBCMD_DROP=                 1,
  UDM_DBCMD_DOCCOUNT=             2,
  UDM_DBCMD_CRAWLER_FINALIZE=     3
} udm_dbcmd_t;


typedef enum
{
  UDM_HREFCMD_START_BULK=         0,
  UDM_HREFCMD_STOP_BULK=          1,
  UDM_HREFCMD_ADD=                2,
  UDM_HREFCMD_EXPORT=             3
} udm_hrefcmd_t;


typedef enum
{
  UDM_DOCCMD_DELETE=              0,
  UDM_DOCCMD_ADD=                 1,
  UDM_DOCCMD_SUPDATE=             2,
  UDM_DOCCMD_LUPDATE=             3,
  UDM_DOCCMD_DUPDATE=             4,
  UDM_DOCCMD_UPDCLONE=            5,
  UDM_DOCCMD_REGCHILD=            6,
  UDM_DOCCMD_FINDBYURL=           7,
  UDM_DOCCMD_FINDBYMSG=           8,
  UDM_DOCCMD_FINDORIG=            9,
  UDM_DOCCMD_GET_CACHED_COPY=    10,
  UDM_DOCCMD_DOCPERSITE=         11,
  UDM_DOCCMD_SQLIMPORTSEC=       12,
  UDM_DOCCMD_RESTOREDATA=        13 
} udm_doccmd_t;


typedef enum
{
  UDM_QUERYCMD_FINDWORDS=           0,
  UDM_QUERYCMD_SUGGEST=             1,
  UDM_QUERYCMD_DOCINFO=             2,
  UDM_QUERYCMD_TRACK=               3,
  UDM_QUERYCMD_CLONES=              4,
  UDM_QUERYCMD_WORDFORMS=           5,
  UDM_QUERYCMD_TARGETS=             6,
  UDM_QUERYCMD_CLEAR=               7,
  UDM_QUERYCMD_INDEX=               8,
  UDM_QUERYCMD_EXPIRE=              9,
  UDM_QUERYCMD_REFERERS=           10,
  UDM_QUERYCMD_DUMPDATA=           11,
  UDM_QUERYCMD_REWRITE_URLDATA=    12,
  UDM_QUERYCMD_REWRITE_LIMITS=     13,
  UDM_QUERYCMD_REWRITE_POPULARITY= 14,
  UDM_QUERYCMD_EXPORT=             15,
  UDM_QUERYCMD_WORDSTAT=           16,
  UDM_QUERYCMD_STATISTICS=         17
} udm_querycmd_t;


typedef enum
{
  UDM_SRV_ACTION_TABLE=           0,
  UDM_SRV_ACTION_FLUSH=           1,
  UDM_SRV_ACTION_ADD=             2
} udm_srvcmd_t;


typedef enum
{
  UDM_DBINFO_ERRCODE=             0,
  UDM_DBINFO_ERRSTR=              1,
  UDM_DBINFO_ADDR=                2,
  UDM_DBINFO_IS_THREAD_SAFE=      3
} udm_dbinfo_t;


UDM_API(void) UdmDBListInit(UDM_DBLIST *List);
UDM_API(udm_rc_t) UdmDBListAdd(UDM_DBLIST *List,
                               const char * addr,
                               char *err, size_t errlen);
void UdmDBListFree(UDM_DBLIST *List);
UDM_API(void) UdmDBInit(UDM_DB *db);
void UdmDBFree(UDM_DB *db);
udm_rc_t UdmDBSetAddr(UDM_DB *db, const char *dbaddr);
int UdmDBIsActive(UDM_AGENT *A, size_t num_arg);


UDM_API(udm_rc_t) UdmDocAction(UDM_AGENT *A, UDM_DOCUMENT   *D, udm_doccmd_t cmd);
UDM_API(udm_rc_t) UdmSrvAction(UDM_AGENT *A, UDM_SERVERLIST *S, udm_srvcmd_t cmd);
UDM_API(udm_rc_t) UdmQueryAction(UDM_AGENT *A, UDM_QUERY *Q, udm_querycmd_t cmd);
UDM_API(udm_rc_t) UdmFind(UDM_AGENT *, UDM_RESULT *);
UDM_API(udm_rc_t) UdmFind2(UDM_AGENT *, UDM_RESULT *, const char *query);
UDM_API(udm_rc_t) UdmDBAction(UDM_AGENT *A, udm_dbcmd_t cmd);
UDM_API(udm_rc_t) UdmEnvFind(UDM_ENV *Env, UDM_RESULT *Res, const char *query);

udm_rc_t UdmHTDBGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc);

size_t UdmDBNumBySeed(UDM_ENV *Env, udmhash32_t seed);


/*******************************/
size_t UdmEnvCopyErrMsgFromDB(UDM_ENV *Env, UDM_DB *db);

typedef struct udm_dbhandler_st
{
  udm_rc_t (*Init)          (UDM_DB *db, const char *addr);
  udm_rc_t (*Close)         (UDM_DB *db);
  udm_rc_t (*Info)          (UDM_DB *db, void *dst, size_t dstlen, size_t *bytes_written, udm_dbinfo_t info);
  udm_rc_t (*QueryAction)   (UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, udm_querycmd_t cmd);
  udm_rc_t (*DBAction)      (UDM_AGENT *A, UDM_DB *db, udm_dbcmd_t cmd);
  udm_rc_t (*DocumentAction)(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *D, udm_doccmd_t cmd);
  udm_rc_t (*HrefAction)    (UDM_AGENT *A, UDM_DB *db, UDM_HREF *H, udm_hrefcmd_t cmd);
  udm_rc_t (*ServerAction)  (UDM_AGENT *A, UDM_DB *db, UDM_SERVERLIST *Srv, udm_srvcmd_t cmd);
} UDM_DBHANDLER;

extern const UDM_DBHANDLER udm_dbhandler_sql;
extern const UDM_DBHANDLER udm_dbhandler_searchd;

#endif
