/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_HASH_H
#define _UDM_HASH_H

#include "udm_config.h"

/* Returns unsigned  hash64 of data block */
udmhash64_t UdmHash64(const char * buf, size_t size);
UDM_API(udmhash32_t) UdmHash32(const char * buf, size_t size);

#define UdmStrHash64(buf)		UdmHash64((buf), strlen(buf))
#define UdmStrHash32(buf)		UdmHash32((buf), strlen(buf))


/*
  "cmp" can be NULL in case if hash value guarantees unique key.
*/
struct udm_hash_st;
typedef struct udm_hash_handler_st
{
  udm_rc_t   (*store)   (struct udm_hash_st *hash, void *key, void *rec);
  udm_rc_t   (*join)    (struct udm_hash_st *hash, void *key, void *rec);
  int        (*cmp)     (struct udm_hash_st *hash, void *key, void *rec);
  udmcrc32_t (*keyhash) (struct udm_hash_st *hash, const void *key);
  udmcrc32_t (*rechash) (struct udm_hash_st *hash, const void *rec);
} UDM_HASH_HANDLER;


typedef struct udm_hash_st
{
  void *base;
  void *user_data;
  size_t mitems;
  size_t nitems;
  size_t item_size;
  UDM_HASH_HANDLER handler;
} UDM_HASH;


udm_rc_t UdmHashInit(UDM_HASH *hash, const UDM_HASH_HANDLER *handler,
                     void *user_data, size_t mitems, size_t item_size);
void UdmHashFree(UDM_HASH *hash);
size_t UdmHashSize(size_t n);
void *UdmHashPut(UDM_HASH *hash, void *ptr);
void *UdmHashFind(UDM_HASH *hash, void *ptr);
size_t UdmHashUsedMemorySize(const UDM_HASH *hash);

#endif
