/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_HTTP_H
#define _UDM_HTTP_H

udm_content_type_t UdmContentTypeByName(const char *str);

size_t UdmHTTPFindContent(const char *src, size_t srclen);
void UdmParseHTTPResponse(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc);
void UdmHTTPBufParseHeadersForCachedCopy(UDM_VARLIST *Vars, const char *src, size_t len);
udm_rc_t UdmStartHTTPD(UDM_AGENT *A, udm_rc_t (*routine)(int sock, UDM_AGENT *A));

const char *UdmHTTPBufPtr(const UDM_HTTPBUF *Buf);
const char *UdmHTTPBufContent(const UDM_HTTPBUF *Buf);
udm_bool_t UdmHTTPBufContentToConstStr(const UDM_HTTPBUF *Buf, UDM_CONST_STR *str);
size_t UdmHTTPBufContentSize(const UDM_HTTPBUF *Buf);
size_t UdmHTTPBufHeaderSize(const UDM_HTTPBUF *Buf);
size_t UdmHTTPBufSize(const UDM_HTTPBUF *Buf);
size_t UdmHTTPBufAllocedSize(const UDM_HTTPBUF *Buf);
size_t UdmHTTPBufAvailableSize(const UDM_HTTPBUF *Buf);
udmcrc32_t UdmHTTPBufCRC32(const UDM_HTTPBUF *Buf);
void UdmHTTPBufContentExport(const UDM_HTTPBUF *Buf, char *str);

void UdmHTTPBufInit(UDM_HTTPBUF *Buf);
void UdmHTTPBufReset(UDM_HTTPBUF *Buf);
void UdmHTTPBufPutContent(UDM_HTTPBUF *Buf, const char *src, size_t len);
void UdmHTTPBufResetContent(UDM_HTTPBUF *Buf);
udm_rc_t UdmHTTPBufAlloc(UDM_HTTPBUF *Buf, size_t size);
udm_rc_t UdmHTTPBufRealloc(UDM_HTTPBUF *Buf, size_t size);
udm_rc_t UdmHTTPBufShrink(UDM_HTTPBUF *Buf);
void UdmHTTPBufFree(UDM_HTTPBUF *Buf);
void UdmHTTPBufTruncateContent(UDM_HTTPBUF *Buf);
udm_rc_t UdmHTTPBufAppend(UDM_HTTPBUF *Buf, const char *src, size_t len);
udm_rc_t UdmHTTPBufAppendHTTPTime(UDM_HTTPBUF *Buf, time_t tm);
udm_rc_t UdmHTTPBufAppendf(UDM_HTTPBUF *Buf, const char *fmt, ...);
udm_rc_t UdmHTTPBufAppendFromFile(UDM_HTTPBUF *Buf, int fd);
int  UdmHTTPBufAppendRecv(UDM_HTTPBUF *Buf, int fd, size_t len, int flags);
udm_rc_t UdmHTTPBufPrintf(UDM_HTTPBUF *Buf, const char *fmt, ...);
void UdmHTTPBufShiftLeft(UDM_HTTPBUF *Buf, size_t offs);
size_t UdmHTTPBufFindContent(UDM_HTTPBUF *Buf);
udm_rc_t UdmHTTPBufInflate(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufInflateAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufDeflateAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufUnGzipAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufUncompressAppend(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufAppendBase64Decode(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufQuotedPrintableDecode(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufAppendQuotedPrintableDecode(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufBase64Decode(UDM_HTTPBUF *Buf, const char *src, size_t srclen);
udm_rc_t UdmHTTPBufAppendEncodingDecode(UDM_HTTPBUF *Buf, const char *src, size_t len, udm_content_encoding_t encoding);
#endif
