/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_LEX_H
#define _UDM_LEX_H


typedef enum 
{
  UDM_LEX_EXCLAM=          '!',        /* 0x21 */
  UDM_LEX_HASH=            '#',        /* 0x23 */
  UDM_LEX_DOLLAR=          '$',        /* 0x24 */
  UDM_LEX_PERCENT=         '%',        /* 0x25 */
  UDM_LEX_AMPERSAND=       '&',        /* 0x26 */
  UDM_LEX_LP=              '(',        /* 0x28 */
  UDM_LEX_RP=              ')',        /* 0x29 */
  UDM_LEX_ASTERISK=        '*',        /* 0x2A */
  UDM_LEX_PLUS=            '+',        /* 0x2B */
  UDM_LEX_COMMA=           ',',        /* 0x2C */
  UDM_LEX_MINUS=           '-',        /* 0x2D */
  UDM_LEX_DOT=             '.',        /* 0x2E */
  UDM_LEX_SLASH=           '/',        /* 0x2F */
  UDM_LEX_COLON=           ':',        /* 0x3A */
  UDM_LEX_SEMICOLON=       ';',        /* 0x3B */
  UDM_LEX_LT=              '<',        /* 0x3C */
  UDM_LEX_EQ=              '=',        /* 0x3D */
  UDM_LEX_GT=              '>',        /* 0x3E */
  UDM_LEX_QUESTION=        '?',        /* 0x3F */
  UDM_LEX_AT=              '@',        /* 0x40 */
  UDM_LEX_LB=              '[',        /* 0x5B */
  UDM_LEX_RB=              ']',        /* 0x5D */
  UDM_LEX_CARET=           '^',        /* 0x5E */
  UDM_LEX_LCB=             '{',        /* 0x7B */
  UDM_LEX_VBAR=            '|',        /* 0x7C */
  UDM_LEX_RCB=             '}',        /* 0x7D */
  UDM_LEX_TILDE=           '~',        /* 0x7E */

#if UDM_EXTENDED_PUNCTUATION_NOT_USED
  UDM_LEX_SPACE=           ' ',        /* 0x20 */
  UDM_LEX_DOUBLE_QUOTE=    '"',        /* 0x22 */  /* Part of a double-quoted string   */
  UDM_LEX_QUOTE=           '\'',       /* 0x27 */  /* Part of a quoted string          */
  UDM_LEX_BACKSLASH=       '\\',       /* 0x5C */  /* Part of a [double] quoted string */
  UDM_LEX_UNDERSCORE=      '_',        /* 0x5F */  /* Part of identifier               */
  UDM_LEX_BACKTICK=        '`',        /* 0x60 */  /* Part of a back-ticked identifier */
#endif

  UDM_LEX_COMMENT,
  UDM_LEX_INC,                         /* ++ */
  UDM_LEX_DEC,                         /* -- */
  UDM_LEX_MUL_EQ,                      /* *= */
  UDM_LEX_DIV_EQ,                      /* /= */
  UDM_LEX_REM_EQ,                      /* %= */
  UDM_LEX_INC_EQ,                      /* += */
  UDM_LEX_DEC_EQ,                      /* -= */
  UDM_LEX_AND_EQ,                      /* &= */
  UDM_LEX_OR_EQ,                       /* |= */
  UDM_LEX_XOR_EQ,                      /* &= */
  UDM_LEX_LE,                          /* <= */
  UDM_LEX_GE,                          /* >= */
  UDM_LEX_EQ_EQ,                       /* == */
  UDM_LEX_NOT_EQ,                      /* != */
  UDM_LEX_LSHIFT_EQ,                   /* <<= */
  UDM_LEX_RSHIFT_EQ,                   /* >>= */

  UDM_LEX_BOOL_OR,                     /* || */
  UDM_LEX_BOOL_AND,                    /* && */

  UDM_LEX_LSHIFT,                      /* << */
  UDM_LEX_RSHIFT,                      /* >> */

  /* C data types */
  UDM_LEX_CHAR,
  UDM_LEX_DOUBLE,
  UDM_LEX_FLOAT,
  UDM_LEX_INT,
  UDM_LEX_LONG,
  UDM_LEX_SHORT,
  UDM_LEX_SIGNED,
  UDM_LEX_UNSIGNED,

  /* C keywords */
  UDM_LEX_AUTO,
  UDM_LEX_CASE,
  UDM_LEX_CONST,
  UDM_LEX_DEFAULT,
  UDM_LEX_ENUM,
  UDM_LEX_EXTERN,
  UDM_LEX_GOTO,
  UDM_LEX_REGISTER,
  UDM_LEX_RETURN,
  UDM_LEX_STATIC,
  UDM_LEX_STRUCT,
  UDM_LEX_SWITCH,
  UDM_LEX_TYPEDEF,
  UDM_LEX_UNION,
  UDM_LEX_VOLATILE,

  UDM_LEX_IF,                          /* if */
  UDM_LEX_ELSE,                        /* else */
  UDM_LEX_WHILE,                       /* while */
  UDM_LEX_FOR,                         /* for */
  UDM_LEX_DO,                          /* do */
  UDM_LEX_BREAK,                       /* break */
  UDM_LEX_CONTINUE,                    /* continue */
  UDM_LEX_SIZEOF,                      /* sizeof */

  /* C++ keywords */
  UDM_LEX_OPERATOR,

  UDM_LEX_COUT,                        /* cout */

  UDM_LEX_CHAR_LITERAL,
  UDM_LEX_STRING=          's',
  UDM_LEX_ESCAPED_STRING=  'S',
  UDM_LEX_IDENT=           'I',
  UDM_LEX_TEXT=            'T',
  UDM_LEX_UINT=            'u',
  UDM_LEX_UNKNOWN=         'U',
  UDM_LEX_EOF=             'E'

} udm_lex_t;


const char *UdmLex2str(udm_lex_t lex);


typedef struct lex_token_st
{
  UDM_CONST_TOKEN token;
  udm_lex_t type;
} UDM_LEX_TOKEN;


typedef struct lex_scanner_st
{
  UDM_LEX_TOKEN token;     /* The last scanned token, for look-ahead */
  UDM_CONST_TOKEN content; /* The entire text to parse      */
  const char *cur;         /* Current position in "content" */
} UDM_LEX_SCANNER;

#define UdmLexScannerEOF(s)    ((s)->cur >= (s)->content.end)
#define UdmLexScannerCur(s)    ((s)->cur)
#define UdmLexScannerShift(s)  ((s)->cur++)
#define UdmLexScannerShiftN(p,n) do { ((p)->cur+= (n)); } while (0)
#define UdmLexScannerStr(s)    ((s)->content.str)
#define UdmLexScannerEnd(s)    ((s)->content.end)
#define UdmLexScannerIsSpace(x)  strchr(" \t\r\n", (x))

void UdmLexScannerSkipSpaces(UDM_LEX_SCANNER *s);
void UdmLexScannerInit(UDM_LEX_SCANNER *s, const char *str, size_t length);
udm_bool_t UdmLexScannerScanString(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanEscapedString(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanChar(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanXMLIdentifier(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanPunctuation(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanIdentifier(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanUnsignedNumber(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanCOperator(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanCComment(UDM_LEX_SCANNER *s, UDM_LEX_TOKEN *a);
udm_bool_t UdmLexScannerScanUntil(UDM_LEX_SCANNER *scanner,
                                  const UDM_CONST_STR *str,
                                  udm_lex_t on_success,
                                  udm_lex_t on_error);
#endif /* UDM_LEX_H */
