/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#ifndef _UDM_VARS_H
#define _UDM_VARS_H

void UdmSectionParamInit(UDM_SECTION_PARAM *Param, udm_secno_t secno);
void UdmSectionParamInitFromVar(UDM_SECTION_PARAM *Param, const UDM_VAR *Var);
udm_rc_t UdmSectionAppendStrn(UDM_SECTION *Section,
                              const char *str, size_t length);
udm_rc_t UdmSectionAppendStrnWithLimitIfFits(UDM_SECTION *Sec,
                                             const UDM_ENV *Env,
                                             const char *str, size_t length);
udm_rc_t
UdmSectionAppendStrnWithLimitFittingPart(UDM_SECTION *Sec,
                                         const UDM_ENV *Env,
                                         const char *str, size_t length);
udm_rc_t
UdmSectionAppendConv(UDM_SECTION *Sec, UDM_CONV *cnv, int cnvflags,
                     const char *src, size_t srclen);
const char *UdmSectionPtr(const UDM_SECTION *Sec);
size_t UdmSectionLength(const UDM_SECTION *Sec);
void UdmSectionSetNullStr(UDM_SECTION *Sec);

UDM_API(void) UdmVarListInit(UDM_VARLIST * vars);
UDM_API(void) UdmVarListFree(UDM_VARLIST * vars);
void UdmVarFree(UDM_VAR *);
void UdmVarDelete(UDM_VAR *);
udm_bool_t UdmVarIsReadOnly(const UDM_VAR *Var);
const char *UdmVarName(const UDM_VAR *Var);
char *UdmVarNameVar(UDM_VAR *Var);
size_t UdmVarNameLength(const UDM_VAR *Var);
udm_secno_t UdmVarSecno(const UDM_VAR *Var);
udm_var_flag_t UdmVarFlags(const UDM_VAR *Var);
udm_value_handler_type_t UdmVarValueHandlerType(const UDM_VAR *);
udm_value_data_type_t UdmVarValueDataType(const UDM_VAR *);
int UdmVarIntVal(const UDM_VAR *);
const char *UdmVarStr(const UDM_VAR *);
size_t UdmVarLength(const UDM_VAR *);
const char *UdmValueStr(const UDM_VALUE *);
UDM_CONST_STR *UdmVarGetConstStr(const UDM_VAR *, UDM_CONST_STR *str);
UDM_SECTION *UdmVarToSection(UDM_VAR *VAR);

udm_rc_t UdmVarListAddEnviron(UDM_VARLIST *vars, const char *name);

udm_rc_t UdmVarAppendStrn(UDM_VAR *var, const char *val, size_t len);
udm_rc_t UdmVarSetStrn(UDM_VAR *var, const char *val, size_t len);
udm_rc_t UdmVarSetConv(UDM_VAR *var, UDM_CONV *conv, int cnvflags,
                         const char *val, size_t len);
udm_rc_t UdmVarAppendConv(UDM_VAR *var, size_t maxlen,
                          UDM_CONV *cnv, int cnvflags,
                          const char *src, size_t srclen);
UDM_API(udm_rc_t) UdmVarListAddStr(UDM_VARLIST * vars,const char * name, const char * val);
udm_rc_t UdmVarListAddStrn(UDM_VARLIST * vars, const char * name, const char * val, size_t len);
udm_rc_t UdmVarListAddQueryStr(UDM_VARLIST *vars, const char *name, const char *val, size_t len);
UDM_API(udm_rc_t) UdmVarListAddQueryString(UDM_VARLIST *Vars, const char *name_prefix, const char *query_string);
udm_rc_t UdmVarListAddQueryStringStrn(UDM_VARLIST *Vars, const char *name_prefix, const char *query_string, size_t length);
udm_rc_t UdmVarListAddInt(UDM_VARLIST * vars,const char * name, int val);
udm_rc_t UdmVarListAddUnsigned(UDM_VARLIST * vars, const char * name, uint4 val);
udm_rc_t UdmVarListAddULongLong(UDM_VARLIST * vars, const char * name, unsigned long long);
udm_rc_t UdmVarListAddDouble(UDM_VARLIST * vars, const char * name, double val);
udm_rc_t UdmVarListDel(UDM_VARLIST * Lst,const char *name);
udm_rc_t UdmVarListDelBySection(UDM_VARLIST * Lst, int secno);
udm_rc_t UdmVarListDelByName(UDM_VARLIST * Lst, const char *name);
udm_rc_t UdmVarCreate(const UDM_VALUE_HANDLER *ha, UDM_VAR **dst,
                      const char *name, const UDM_VALUE **args, size_t nargs);
udm_rc_t UdmVarListAddLst(UDM_VARLIST * Lst, const UDM_VARLIST *Src,
                          const char *name, const char *pattern);
udm_rc_t UdmVarListReplaceLst(UDM_VARLIST * Lst,const UDM_VARLIST *Src,
                              const char *name, const char *pattern);
udm_rc_t UdmVarListMerge(UDM_VARLIST * Lst,UDM_VARLIST *Src1, UDM_VARLIST *Src2);

UDM_API(udm_rc_t) UdmVarListReplaceStr(UDM_VARLIST * vars,const char * name,const char * val);
UDM_API(udm_rc_t) UdmVarListReplaceInt(UDM_VARLIST * vars,const char * name,int val);
UDM_API(udm_rc_t) UdmVarListReplaceUnsigned(UDM_VARLIST * vars, const char * name, uint4 val);
udm_rc_t UdmVarListReplaceULongLong(UDM_VARLIST * vars,const char * name, unsigned long long);
udm_rc_t UdmVarListReplaceDouble(UDM_VARLIST * vars, const char * name, double val);

udm_rc_t UdmVarListReplaceStrn(UDM_VARLIST *vars, const char *name,
                               const char *val, size_t vallen);
udm_rc_t
UdmVarListReplaceOrAppendStrn(UDM_VARLIST *vars, const char *name,
                              const char *val, size_t len, int hl);

udm_rc_t UdmVarListReplaceStrnWithParam(UDM_VARLIST *vars,
                                        const UDM_SECTION_PARAM *Param,
                                        const char *name,
                                        const char *val, size_t vallen);
udm_rc_t UdmVarListAddStrnWithParam(UDM_VARLIST *vars,
                                    const UDM_SECTION_PARAM *Param,
                                    const char *name,
                                    const char *val, size_t vallen);

size_t UdmDSTRParse(UDM_DSTR *d, const char *fmt, const UDM_VARLIST *L);

udm_rc_t UdmVarListInsStr(UDM_VARLIST * vars,const char * name,const char * val);
udm_rc_t UdmVarListInsInt(UDM_VARLIST * vars,const char * name,int val);

UDM_VAR *UdmVarListFindVar(const UDM_VARLIST *vars,const char *name);
UDM_VAR *UdmVarListFindOrCreateStr(UDM_VARLIST *vars, const char *name);

UDM_VAR *UdmVarListFindByIndex(const UDM_VARLIST *Vars, size_t i);
const UDM_VAR *UdmVarListFindConstByIndex(const UDM_VARLIST *Vars, size_t i);

const UDM_VAR *UdmVarListFind(const UDM_VARLIST *vars,const char * name);
const UDM_VAR *UdmVarListFindBySecno(const UDM_VARLIST *vars, int secno);
const UDM_VAR *UdmVarListFindWithValue(const UDM_VARLIST * vars, const char *name,
                                       const char *val, size_t vallength);
const UDM_VAR *UdmVarListFindByPrefix(const UDM_VARLIST *Vars,
                                      const char *prefix, size_t prefix_length);

const UDM_VALUE *UdmVarListFindConstValue(const UDM_VARLIST *Vars, const char *name);
UDM_VALUE *UdmVarListFindValue(const UDM_VARLIST *Vars, const char *name);

UDM_API(const char *)UdmVarListFindStr(const UDM_VARLIST *vars, const char *name, const char * defval);
UDM_API(int)         UdmVarListFindInt(const UDM_VARLIST *vars, const char *name, int defval);
UDM_API(udm_bool_t)  UdmVarListFindBool(const UDM_VARLIST *vars, const char *name, udm_bool_t defval);
UDM_API(unsigned)    UdmVarListFindUnsigned(const UDM_VARLIST *vars, const char *name, unsigned defval);
UDM_API(double)      UdmVarListFindDouble(const UDM_VARLIST *vars, const char *name, double defval);
UDM_API(UDM_CHARSET*)UdmVarListFindCharset(const UDM_VARLIST *vars, const char *name, UDM_CHARSET *def);

const char *UdmVarListFindStrNonEmpty(const UDM_VARLIST *vars, const char *name, const char *defval);
udm_content_type_t UdmVarListFindContentType(const UDM_VARLIST *Lst, const char *name, udm_content_type_t def);
udm_bool_t UdmVarListFindConstStr(UDM_CONST_STR *buf,
                                  const UDM_VARLIST *vars,
                                  const char *name,
                                  const UDM_CONST_STR *defval);
udm_rc_t UdmVarListConvert(UDM_VARLIST *L, UDM_CONV *conv, int flags);
size_t UdmVarListUsedMemory(const UDM_VARLIST *L);
void UdmVarListPrint(const UDM_VARLIST *L, FILE *f);

void UdmFeatures(UDM_VARLIST *L);

udm_rc_t UdmSectionListPrint(const UDM_VARLIST *L, FILE *f);
udm_rc_t UdmVarListReplaceVar(UDM_VARLIST *Vars, UDM_VAR *Var);

void UdmVarFreeGeneric(UDM_VAR *Var);
extern UDM_VALUE_HANDLER UdmValueHandlerSimple;
extern UDM_VALUE_HANDLER UdmValueHandlerInt;
extern UDM_VALUE_HANDLER UdmValueHandlerDouble;
extern UDM_VALUE_HANDLER UdmValueHandlerChar;
extern UDM_VALUE_HANDLER UdmValueHandlerEnv;
extern UDM_VALUE_HANDLER UdmValueHandlerResult;
extern UDM_VALUE_HANDLER UdmValueHandlerDocument;
extern UDM_VALUE_HANDLER UdmValueHandlerSQLResult;
extern UDM_VALUE_HANDLER UdmValueHandlerExcerptFragment;
extern UDM_VALUE_HANDLER UdmValueHandlerQueryWord;

void UdmVarListListInit(UDM_VARLISTLIST *List);
void UdmVarListListFree(UDM_VARLISTLIST *List);
udm_rc_t UdmVarListListAdd(UDM_VARLISTLIST *List);
udm_rc_t UdmVarListListAddWithFlags(UDM_VARLISTLIST *List, udm_varlist_flag_t);
void UdmVarListListLastDelete(UDM_VARLISTLIST *List);

UDM_VALUE *UdmValueNew(const UDM_VALUE_HANDLER *ha);
void UdmValueDelete(UDM_VALUE *Value);

char *UdmValueDataPtr(UDM_VALUE *Value);
const char *UdmValueConstDataPtr(const UDM_VALUE *Value);
char *UdmVarDataPtr(UDM_VAR *);
const char *UdmVarConstDataPtr(const UDM_VAR *);

#endif
