/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <math.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h" /* For UdmValueHandlerXXX */
#include "udm_prog.h"


static void
UdmFunctionSQRT(UDM_PROG_EXECUTOR_STATE *state)
{
  double x= ((double*) state->vreg[0].Data)[0];
  UDM_FUNC_RETURN_DOUBLE(state, sqrt(x));
}


static void
UdmFunctionGetEnv(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  char *str, name[128];
  udm_snprintf(name, sizeof(name), "%.*s",
               (int) UdmDSTRLength(&Section->Value),
               UdmDSTRPtr(&Section->Value));
  str= getenv(name);
  UdmProgExecutorStateReturnStrn(state, str, str ? strlen(str) : 0);
}


static void
UdmFunction_argv(UDM_PROG_EXECUTOR_STATE *state)
{
  int argno= ((int *) state->vreg[0].Data)[0];
  const char *str= argno <= state->globals.argc ?
                   state->globals.argv[argno] : "";
  UdmProgExecutorStateReturnStrn(state, str, str ? strlen(str) : 0);
}


static void
UdmFunction_int_to_string(UDM_PROG_EXECUTOR_STATE *state)
{
  int *src= (int *) state->vreg[0].Data;
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UDM_OK != UdmDSTRAppendf(&ret->Value, "%d", *src);
}


static void
UdmFunction_double_to_string(UDM_PROG_EXECUTOR_STATE *state)
{
  double *src= (double *) state->vreg[0].Data;
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UDM_OK != UdmDSTRAppendf(&ret->Value, "%f", *src);
}


static void
UdmFunctionVersion(UDM_PROG_EXECUTOR_STATE *state)
{
  UdmProgExecutorStateReturnStrn(state, VERSION, strlen(VERSION));
}


static void
UdmFunction_rand(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_FUNC_RETURN_INT(state, rand());
}


static void
UdmFunction_time(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_FUNC_RETURN_INT(state, time(NULL));
}


static void
UdmFunction_srand(UDM_PROG_EXECUTOR_STATE *state)
{
  int x= ((int*) state->vreg[0].Data)[0];
  srand(x);
}


static void
UdmFunction_argc(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_FUNC_RETURN_INT(state, state->globals.argc);
}


void
UdmFunction_exit(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_FUNC_RETURN_INT(state, ((int*) state->vreg[0].Data)[0]);
}


static UDM_VALUE_PROTOTYPE prototype_int[]= {{&UdmValueHandlerInt}};
static UDM_VALUE_PROTOTYPE prototype_double[]= {{&UdmValueHandlerDouble}};
static UDM_VALUE_PROTOTYPE prototype_string[]= {{&UdmValueHandlerSimple}};


const UDM_FUNCTION udm_builtin_functions[]=
{
  {{UDM_CSTR_WITH_LEN("sqrt")},
    {{&UdmValueHandlerDouble}, 1, prototype_double}, UdmFunctionSQRT},
  {{UDM_CSTR_WITH_LEN("getenv")},
    {{&UdmValueHandlerSimple}, 1, prototype_string}, UdmFunctionGetEnv},
  {{UDM_CSTR_WITH_LEN("_argv")},
    {{&UdmValueHandlerSimple}, 1, prototype_int}, UdmFunction_argv},
  {{UDM_CSTR_WITH_LEN("mnogosearch_version")},
    {{&UdmValueHandlerSimple}, 0, NULL}, UdmFunctionVersion},
  {{UDM_CSTR_WITH_LEN("to_string")},
    {{&UdmValueHandlerSimple}, 1, prototype_int}, UdmFunction_int_to_string},
  {{UDM_CSTR_WITH_LEN("to_string")},
    {{&UdmValueHandlerSimple}, 1, prototype_double}, UdmFunction_double_to_string},
  {{UDM_CSTR_WITH_LEN("rand")},
    {{&UdmValueHandlerInt}, 0, NULL}, UdmFunction_rand},
  {{UDM_CSTR_WITH_LEN("time")},
    {{&UdmValueHandlerInt}, 0, NULL}, UdmFunction_time},
  {{UDM_CSTR_WITH_LEN("srand")},
    {{NULL}, 1, prototype_int}, UdmFunction_srand},
  {{UDM_CSTR_WITH_LEN("exit")},
    {{NULL}, 1, prototype_int}, UdmFunction_exit},
  {{UDM_CSTR_WITH_LEN("_argc")},
    {{&UdmValueHandlerInt}, 0, NULL}, UdmFunction_argc},

  {{NULL, 0}, {{NULL}, 0, NULL}, NULL}, 
};
