/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <string.h>

#include "udm_common.h"
#include "udm_match.h"
#include "udm_http.h"
#include "udm_textlist.h"
#include "udm_vars.h"
#include "udm_parsehtml.h"
#include "udm_indexer.h"

typedef struct udm_parse_section_helper_st
{
  UDM_AGENT *Indexer;
  UDM_USERSECTIONLIST *SectionMatch;
  UDM_DOCUMENT *Doc;
  UDM_CHARSET *doccs;
  UDM_DSTR tbuf;
  UDM_TEXTITEM Item;
} UDM_PARSE_SECTION_HELPER;


static udm_rc_t
UdmParseSectionsHTMLText(UDM_HTML_PARSER *parser)
{
  UDM_PARSE_SECTION_HELPER *param= (UDM_PARSE_SECTION_HELPER *) parser->user_data;
  if (!parser->state.script && !parser->state.comment && !parser->state.style)
  {
    UdmDSTRReset(&param->tbuf);
    UdmDSTRAppend(&param->tbuf, parser->lasttok.str, parser->lasttok.length);
    param->Item.str= param->tbuf.Val.str;
    if (&param->Indexer->Conf->SectionHdrMatch == param->SectionMatch)
    {
      UdmVarListReplaceStr(&param->Doc->Sections,
                           param->Item.section_name, param->Item.str);
    }
    else if (&param->Indexer->Conf->SectionGsrMatch == param->SectionMatch &&
             param->doccs)
    {
      char *trmsrc= UdmTrim(param->Item.str, " \t\r\n");
      size_t trmlen= strlen(trmsrc);
      size_t cnvlen= trmlen * 12 + 1;
      char *cnvsrc= (char*) UdmMalloc(cnvlen);
      UdmVarListReplaceStr(&param->Doc->Sections,
                           param->Item.section_name, param->Item.str);
      if (trmsrc[0] && cnvsrc)
      {
        UDM_CONV lcs_dcs;
        UdmConvInit(&lcs_dcs, param->Doc->lcs, param->doccs);
        UdmConvHTML(&lcs_dcs, cnvsrc, cnvlen, trmsrc, trmlen + 1);
        param->Item.str= cnvsrc;
        param->Item.Param.flags= (udm_textlist_flag_t)
          (param->Item.Param.flags | UDM_TEXTLIST_FLAG_SKIP_ADD_SECTION);
        UdmTextListAdd(&param->Doc->TextList, &param->Item);
        UdmFree(cnvsrc);
      }
    }
    else
    {
      UdmTextListAdd(&param->Doc->TextList, &param->Item);
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmParseSectionsHTML(UDM_PARSE_SECTION_HELPER *param, const char *src, size_t srclen)
{
  UDM_HTML_PARSER parser;
  UdmHTMLParserInit(&parser);
  UdmHTMLParserSetUserData(&parser, param);
  UdmHTMLParserSetTextHandler(&parser, UdmParseSectionsHTMLText);
  return UdmHTMLParserExec(&parser, src, srclen);
}


/*
  Parse user defined sections
*/
udm_rc_t
UdmUserSectionsParse(UDM_AGENT *Indexer,
                     UDM_USERSECTIONLIST *SectionMatch,
                     UDM_DOCUMENT *Doc, UDM_CHARSET *doccs)
{
  size_t i;
  UDM_PARSE_SECTION_HELPER param;
  UDM_DSTR sbuf;

  if (!SectionMatch->nitems)
    return UDM_OK;
  
  /* Init user data */
  UdmTextItemInit(&param.Item, UDM_TEXTLIST_FLAG_HTML); /* TODO34: indexer.conf: format */
  UdmDSTRInit(&param.tbuf, 1024);
  param.Indexer= Indexer;
  param.SectionMatch= SectionMatch;
  param.Doc= Doc;
  param.doccs= doccs;

  UdmDSTRInit(&sbuf, 1024);

  for (i = 0; i < SectionMatch->nitems; i++)
  {
    UDM_VAR *Sec;
    UDM_MATCH_PART  Parts[10];
    size_t buflen, nparts= 10;
    UDM_USERSECTION *Alias= &SectionMatch->Item[i];
    UDM_CONST_STR content;

    if (Alias->Source.str)
    {
      UdmDSTRReset(&param.tbuf);
      UdmDSTRParse(&param.tbuf, Alias->Source.str, &Doc->Sections);
      content.str= UdmDSTRPtr(&param.tbuf);
      content.length= UdmDSTRLength(&param.tbuf);
    }
    else if (UdmHTTPBufContentToConstStr(&Doc->Buf, &content)) /* TODO34: UdmHTTPBuf: C-String: net_string arg to UdmMatchExec */
      continue; /* No content? */

    if (!(Sec= UdmVarListFindVar(&Doc->Sections, Alias->SectionName.str)))
      continue;

    if (UdmMatchExec(&Alias->Replace.Match, content.str, content.length,
                     content.str, nparts, Parts))
      continue;

    /* Calculate space required for UdmMatchApply, and allocate space */
    buflen= UdmMatchApply(NULL, 1, content.str,
                          Alias->Replace.Replace.str,
                          Alias->Replace.Replace.length,
                          &Alias->Replace.Match, nparts, Parts);
    UdmDSTRReset(&sbuf);
    UdmDSTRRealloc(&sbuf, buflen);

    if (!(sbuf.Val.length= UdmMatchApply(sbuf.Val.str, buflen, content.str,
                                         Alias->Replace.Replace.str,
                                         Alias->Replace.Replace.length,
                                         &Alias->Replace.Match,
                                         nparts, Parts)))
      continue;

    UDM_ASSERT(UdmDSTRLength(&sbuf) + 1 == buflen);/* TODO34!!! why + 1?*/

    param.Item.section_name= UdmVarNameVar(Sec);
    /* TODO34: indexer.conf: fmt*/
    UdmTextParamInit(&param.Item.Param, UDM_TEXTLIST_FLAG_HTML, UdmVarSecno(Sec));
    UdmParseSectionsHTML(&param, UdmDSTRPtr(&sbuf), UdmDSTRLength(&sbuf));
  }

  UdmDSTRFree(&sbuf);

  /* Free user data */
  UdmDSTRFree(&param.tbuf);
  return UDM_OK;
}


udm_rc_t
UdmUserSectionListAdd(UDM_USERSECTIONLIST *L,
                      const UDM_MATCH_PARAM *MatchParam,
                      const UDM_CONST_STR *Section,
                      const UDM_CONST_STR *Source,
                      const UDM_CONST_STR *Pattern,
                      const UDM_CONST_STR *Replacement,
                      char *err, size_t errsize)
{
  UDM_USERSECTION *N;

  UDM_ASSERT(Section->str);
  UDM_ASSERT(Pattern->str);
  UDM_ASSERT(Replacement->str);
  if (L->nitems >= L->mitems)
  {
    L->mitems= L->nitems + 32;
    L->Item= (UDM_USERSECTION*) UdmRealloc(L->Item, L->mitems * sizeof(UDM_USERSECTION));
  }

  N= &L->Item[L->nitems++];
  UdmMatchInit(&N->Replace.Match);
  UdmMatchSetPattern(&N->Replace.Match, Pattern->str);
  N->Replace.Match.Param= *MatchParam;
  N->SectionName.str= udm_strndup(Section->str, Section->length);
  N->SectionName.length= Section->length;
  N->Replace.Replace.str= udm_strndup(Replacement->str, Replacement->length);
  N->Replace.Replace.length= Replacement->length;
  N->Source.str= Source->str ? udm_strndup(Source->str, Source->length) : NULL;
  N->Source.length= Source->length;
  return UdmMatchComp(&N->Replace.Match, err, errsize);
}


void
UdmUserSectionListInit(UDM_USERSECTIONLIST *L)
{
  bzero((void*)L, sizeof(*L));
}


void
UdmUserSectionListFree(UDM_USERSECTIONLIST *L)
{
  size_t i;
  for (i= 0; i < L->nitems; i++)
  {
    UdmMatchFree(&L->Item[i].Replace.Match);
    UDM_FREE(L->Item[i].Source.str);
    UDM_FREE(L->Item[i].SectionName.str);
    UDM_FREE(L->Item[i].Replace.Replace.str);
  }
  L->nitems= L->mitems= 0;
  UDM_FREE(L->Item);
}


const UDM_USERSECTION *
UdmUserSectionListFind(const UDM_USERSECTIONLIST *L, const char *str,
                       size_t nparts, UDM_MATCH_PART *Parts)
{
  size_t i, slen= strlen(str);
  for (i= 0; i < L->nitems; i++)
  {
    UDM_USERSECTION *S= &L->Item[i];
    if (!UdmMatchExec(&S->Replace.Match, str, slen, str, nparts, Parts))
      return S;
  }
  return NULL;
}
