/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/


#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_result.h"
#include "udm_db.h"
#include "udm_env.h"
#include "udm_prog.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_RESULT Result;
} UDM_VALUE_RESULT;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_RESULT Value;
} UDM_VAR_RESULT;


static udm_rc_t
UdmValueResultDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueResultFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueResultSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueResultMemUsed(const char *Data)
{
  return sizeof(UDM_RESULT);
}


static void
UdmValueResultDestructor(char *Data)
{
  UDM_RESULT *Res= (UDM_RESULT *) Data;
  UdmResultFree(Res);
}


static udm_rc_t
UdmValueResultCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  memmove(D, S, sizeof(UDM_RESULT));
  return UDM_OK;
}


static udm_rc_t
UdmValueResultConstructor(char *Data,
                          const UDM_VALUE **args, size_t nargs)
{
  if (args && args[0] && args[0]->handler->datatype == UDM_VALUE_DATA_TYPE_ENV)
  {
    UDM_RESULT *Result;
    if (!(Result= (UDM_RESULT *) UdmMalloc(sizeof(UDM_RESULT))))
      return UDM_ERROR;
    UdmResultInit(Result);
    ((UDM_RESULT *) Data)[0]= Result[0];
  }
  else
  {
    UDM_BZERO(Data, sizeof(UDM_RESULT));
  }
  return UDM_OK;
}


static size_t
UdmValueResultVarSize(void)
{
  return sizeof(UDM_VAR_RESULT);
}


static size_t
UdmValueResultValueSize(void)
{
  return sizeof(UDM_VALUE_RESULT);
}


static size_t
UdmValueResultDataOffset(void)
{
  return offsetof(UDM_VALUE_RESULT, Result);
}


static size_t
UdmValueResultDataSize(void)
{
  return sizeof(UDM_RESULT);
}


static size_t
UdmValueResultDataAlignment(void)
{
  return sizeof(size_t);
}


static void
UdmValueResultGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#Result#")};
  *Val= name;
}


static void
UdmValueResultGetBool(const char *Data, udm_bool_t *val)
{
  *val= UDM_TRUE;
}


static void
UdmValueResultGetInt(const char *Data, int *val)
{
  *val= 0;
}


static void
UdmValueResultGetUInt(const char *Data, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueResultGetDouble(const char *Data, double *val)
{
  *val= 0.0;
}


static udm_rc_t
UdmValueResultSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultAppendConv(char *Data, size_t maxlen,
                         UDM_CONV *cnv, int cnvflags,
                         const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueResultPrintToFile(const char *Data, FILE *f)
{
  return UDM_OK;
}


static udm_rc_t
UdmValueResultPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  return UDM_OK;
}


static void
UdmResultMethodFind(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_ENV *Env= ((UDM_ENV **) state->vreg[1].Data)[0];
  const UDM_SECTION *Section= (const UDM_SECTION*) state->vreg[2].Data;
  char str[256];
  Env->errstr[0]= '\0';
  Env->THDHandler= state->globals.THDHandler;
  udm_snprintf(str, sizeof(str), "%.*s",
               (int) UdmDSTRLength(&Section->Value), UdmDSTRPtr(&Section->Value));
  /*
    TODO34: if Env is reused for multiple queries, this UdmEnvPrepate()
    will load ispell multiple times. Fix this to avoid multiple loading.
  */
  if (UDM_FUNC_RETURN_INT(state, UdmEnvPrepare(Env)))
    return; /* Error, e.g. could not load ispell files */
  UDM_FUNC_RETURN_INT(state, UdmEnvFind(Env, Result, str));
}

static UDM_VALUE_PROTOTYPE UdmResultMethodPrototypeFind[]=
{{&UdmValueHandlerResult},{&UdmValueHandlerEnv},{&UdmValueHandlerSimple}};


static void
UdmResultMethodTotalFound(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) UdmResultTotalFound(Result));
}

static void
UdmResultMethodFirst(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) UdmResultFirst(Result));
}

static void
UdmResultMethodLast(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) UdmResultLast(Result));
}

static void
UdmResultMethodNumRows(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) UdmResultNumRows(Result));
}


static void
UdmResultMethodNumWords(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) Result->WWList.nwords);
}


static void
UdmResultMethodNumUniqWords(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT*) state->vreg[0].Data;
  UDM_FUNC_RETURN_INT(state, (int) Result->WWList.nuniq);
}


static udm_rc_t
UdmResultPropertyAppendHilight(UDM_DSTR *dst,
                               UDM_CHARSET *cs,
                               const UDM_CONST_STR *str,
                               const char *hlbeg, size_t hlbeglen,
                               const char *hlend, size_t hlendlen)
{
  UDM_DSTR tmp;
  UDM_CONV cnv;
  UdmConvInit(&cnv, cs, cs);
  UdmDSTRInit(&tmp, 128);
  UdmDSTRAppendConv(&tmp, &cnv, str->str, str->length, UDM_RECODE_HTML);
  UdmDSTRAppendHighlight(dst, UdmDSTRPtr(&tmp), UdmDSTRLength(&tmp),
                         hlbeg, hlbeglen, hlend, hlendlen);
  UdmDSTRFree(&tmp);
  return UDM_OK;
}


static UDM_VALUE *
UdmVarListFindValueByDSTR(UDM_VARLIST *Vars, const UDM_DSTR *str)
{
  char name[128];
  udm_snprintf(name, sizeof(name), "%.*s",
               (int) UdmDSTRLength(str), UdmDSTRPtr(str));
  return UdmVarListFindValue(Vars, name);
}


static void
UdmResultMethodWord(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_RESULT *Result= (UDM_RESULT *) state->vreg[0].Data;
  int pos= ((int *) state->vreg[1].Data)[0];
  void *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerQueryWord);
  if (pos < 0 || (size_t) pos >= Result->WWList.nwords)
    return;
  state->fatal_error= UdmValueHandlerQueryWord.Copy(ret,
                                                   (char *) &Result->WWList.Word[pos]);
}


static void
UdmResultMethodPropertyCommon(UDM_PROG_EXECUTOR_STATE *state,
                              UDM_SECTION *ReturnValue,
                              UDM_VARLIST *Vars, UDM_SECTION *Property,
                              UDM_CHARSET *cs, udm_bool_t html,
                              const char *hlbeg, size_t hlbeglen,
                              const char *hlend, size_t hlendlen)
{
  UDM_CONST_STR str;
  UDM_VALUE *Value;
  if (!(Value= UdmVarListFindValueByDSTR(Vars, &Property->Value)))
    return;
  Value->handler->GetConstStr(UdmValueDataPtr(Value), &str);
  state->fatal_error= html ?
    UdmResultPropertyAppendHilight(&ReturnValue->Value, cs, &str,
                                   hlbeg, hlbeglen,
                                   hlend, hlendlen) :
    UdmDSTRAppendRemoveHiLight(&ReturnValue->Value, str.str, str.length);
}


static void
UdmResultMethodDocumentPropertyInternal(UDM_PROG_EXECUTOR_STATE *state,
                                        udm_bool_t html,
                                        const char *hlbeg, size_t hlbeglen,
                                        const char *hlend, size_t hlendlen)
{
  UDM_RESULT *Result= (UDM_RESULT *) state->vreg[0].Data;
  int pos= ((int *) state->vreg[1].Data)[0];
  UDM_SECTION *Property= (UDM_SECTION *) state->vreg[2].Data;
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  if (pos < 0 || (size_t) pos >= Result->num_rows)
    return;
  UdmResultMethodPropertyCommon(state, ret, &Result->Doc[pos].Sections,
                                       Property,
                                       Result->Doc[pos].lcs, html,
                                       hlbeg, hlbeglen, hlend, hlendlen);
}


static void
UdmResultMethodPropertyInternal(UDM_PROG_EXECUTOR_STATE *state,
                                UDM_CHARSET *cs,
                                udm_bool_t html,
                                const char *hlbeg, size_t hlbeglen,
                                const char *hlend, size_t hlendlen)
{
  UDM_RESULT *Result= (UDM_RESULT *) state->vreg[0].Data;
  UDM_SECTION *Property= (UDM_SECTION *) state->vreg[1].Data;
  UDM_SECTION *ret= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UdmResultMethodPropertyCommon(state, ret, &Result->Vars,
                                Property, cs, html,
                                hlbeg, hlbeglen, hlend, hlendlen);
}


static void
UdmResultMethodDocumentProperty(UDM_PROG_EXECUTOR_STATE *state)
{
  UdmResultMethodDocumentPropertyInternal(state, UDM_FALSE, "", 0, "", 0);
}


static void
UdmResultMethodDocumentPropertyHtml(UDM_PROG_EXECUTOR_STATE *state)
{
  UdmResultMethodDocumentPropertyInternal(state, UDM_TRUE, "", 0, "", 0);
}


static void
UdmResultMethodDocumentPropertyHtml5(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *HlBeg= (UDM_SECTION *) state->vreg[3].Data;
  UDM_SECTION *HlEnd= (UDM_SECTION *) state->vreg[4].Data;
  UdmResultMethodDocumentPropertyInternal(state, UDM_TRUE,
                                          UdmDSTRPtr(&HlBeg->Value),
                                          UdmDSTRLength(&HlBeg->Value),
                                          UdmDSTRPtr(&HlEnd->Value),
                                          UdmDSTRLength(&HlEnd->Value));
}


static void
UdmResultMethodProperty(UDM_PROG_EXECUTOR_STATE *state)
{
  UdmResultMethodPropertyInternal(state, &udm_charset_latin1, UDM_FALSE,
                                  "", 0, "", 0);
}


static UDM_VALUE_PROTOTYPE UdmValuePrototypeThis[]= {{&UdmValueHandlerResult}};


static UDM_VALUE_PROTOTYPE UdmValuePrototypeResultString[]=
{{&UdmValueHandlerResult},
 {&UdmValueHandlerSimple},
};


static UDM_VALUE_PROTOTYPE UdmValuePrototypeResultIntString[]=
{{&UdmValueHandlerResult},
 {&UdmValueHandlerInt},
 {&UdmValueHandlerSimple},
 {&UdmValueHandlerSimple},
 {&UdmValueHandlerSimple},
};


static const UDM_FUNCTION ResultMethods[]=
{
  {{UDM_CSTR_WITH_LEN("find")},
    {{&UdmValueHandlerInt}, 3, UdmResultMethodPrototypeFind},
    UdmResultMethodFind},
  {{UDM_CSTR_WITH_LEN("total_found")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodTotalFound},
  {{UDM_CSTR_WITH_LEN("first")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodFirst},
  {{UDM_CSTR_WITH_LEN("last")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodLast},
  {{UDM_CSTR_WITH_LEN("num_rows")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodNumRows},
  {{UDM_CSTR_WITH_LEN("num_words")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodNumWords},
  {{UDM_CSTR_WITH_LEN("num_uniq_words")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmResultMethodNumUniqWords},
  {{UDM_CSTR_WITH_LEN("document_property")},
    {{&UdmValueHandlerSimple}, 3, UdmValuePrototypeResultIntString},
    UdmResultMethodDocumentProperty},
  {{UDM_CSTR_WITH_LEN("document_property_html")},
    {{&UdmValueHandlerSimple}, 3, UdmValuePrototypeResultIntString},
    UdmResultMethodDocumentPropertyHtml},
  {{UDM_CSTR_WITH_LEN("document_property_html")},
    {{&UdmValueHandlerSimple}, 5, UdmValuePrototypeResultIntString},
    UdmResultMethodDocumentPropertyHtml5},
  {{UDM_CSTR_WITH_LEN("property")},
    {{&UdmValueHandlerSimple}, 2, UdmValuePrototypeResultString},
    UdmResultMethodProperty},
  {{UDM_CSTR_WITH_LEN("word")},
    {{&UdmValueHandlerQueryWord}, 2, UdmValuePrototypeResultIntString},
    UdmResultMethodWord},
  {{0,0}, {{NULL}, 0, NULL}, NULL}
};


UDM_VALUE_HANDLER UdmValueHandlerResult=
{
  UDM_VALUE_HANDLER_TYPE_RESULT,
  UDM_VALUE_DATA_TYPE_RESULT,
  UDM_VALUE_DATA_TYPE_STR,
  "RESULT",
  UdmValueResultConstructor,
  UdmValueResultDestructor,
  UdmValueResultCopy,
  UdmValueResultVarSize,
  UdmValueResultValueSize,
  UdmValueResultDataOffset,
  UdmValueResultDataSize,
  UdmValueResultDataAlignment,
  UdmValueResultDump,
  UdmValueResultFlags,
  UdmValueResultSecno,
  UdmValueResultMemUsed,
  UdmValueResultGetConstStr,
  UdmValueResultGetBool,
  UdmValueResultGetInt,
  UdmValueResultGetUInt,
  UdmValueResultGetDouble,
  UdmValueResultSetInt,
  UdmValueResultSetDouble,
  UdmValueResultSetStrn,
  UdmValueResultPrintToFile,
  UdmValueResultPrintToDSTR,
  UdmValueResultConvertCharset,
  UdmValueResultSetConv,
  UdmValueResultAppendStrn,
  UdmValueResultAppendConv,
  ResultMethods,
};
