/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_searchtool.h"


static void
UdmDSTRAppendComma(UDM_DSTR *dstr)
{
  if (UdmDSTRLength(dstr))
    UdmDSTRAppend(dstr, UDM_CSTR_WITH_LEN(", "));
}


static size_t
UdmWWListSumCount(const UDM_WIDEWORDLIST *WWList, size_t order)
{
  size_t count, j;
  for (count= j= 0; j < WWList->nwords; j++)
  {
    if (WWList->Word[j].Param.order != order)
      continue;
    switch (WWList->Word[j].Param.origin)
    {
    case UDM_WORD_ORIGIN_QUERY:
    case UDM_WORD_ORIGIN_SPELL:
    case UDM_WORD_ORIGIN_SYNONYM:
    case UDM_WORD_ORIGIN_SYNONYM_FINAL:
    case UDM_WORD_ORIGIN_COLLATION:
      count += WWList->Word[j].Param.count;
      break;
    case UDM_WORD_ORIGIN_STOP:
    case UDM_WORD_ORIGIN_SUGGEST:
      break;
    }
  }
  return count;
}


static void
UdmWWListWordInfoShort(const UDM_WIDEWORDLIST *WWList,
                       UDM_DSTR *dstr, int print_own_count)
{
  size_t i;
  for(i= 0; i < WWList->nwords; i++)
  {
    const UDM_WIDEWORD *W= &WWList->Word[i];
    size_t count= UdmWWListSumCount(WWList, W->Param.order);
    if (W->Param.origin == UDM_WORD_ORIGIN_STOP)
    {
      UdmDSTRAppendComma(dstr);
      UdmDSTRAppendf(dstr, "%s : stopword", W->Word.str);
    }
    else if (W->Param.origin == UDM_WORD_ORIGIN_QUERY)
    {
      UdmDSTRAppendComma(dstr);
      if (print_own_count)
        UdmDSTRAppendf(dstr, "%s : %d / %d",
                             W->Word.str,  (int) W->Param.count, (int) count);
      else
        UdmDSTRAppendf(dstr, "%s : %d", W->Word.str, (int) count);
    }
  }
}


static void
UdmWWListWordInfoExtended(const UDM_WIDEWORDLIST *WWList, UDM_DSTR *dstr)
{
  size_t i;
  for(i= 0; i < WWList->nwords; i++)
  {
    UDM_WIDEWORD *W= &WWList->Word[i];
    switch (W->Param.origin)
    {
      case UDM_WORD_ORIGIN_QUERY:
      case UDM_WORD_ORIGIN_SPELL:
      case UDM_WORD_ORIGIN_SYNONYM:
      case UDM_WORD_ORIGIN_SYNONYM_FINAL:
      case UDM_WORD_ORIGIN_COLLATION:
        UdmDSTRAppendComma(dstr);
        UdmDSTRAppendf(dstr, "%s : %d", W->Word.str, (int) W->Param.count);
        break;
      case UDM_WORD_ORIGIN_STOP:
        UdmDSTRAppendComma(dstr);
        UdmDSTRAppendf(dstr, "%s : stopword", WWList->Word[i].Word.str);
        break;
      case UDM_WORD_ORIGIN_SUGGEST:
        break;
    }
  }
}


static void
UdmWWListWordInfoUniqueWords(UDM_VARLIST *Vars, const UDM_WIDEWORDLIST *WWList)
{
  size_t i;
  for (i= 0; i < WWList->nwords; i++)
  {
    UDM_WIDEWORD *W= &WWList->Word[i];
    if (W->Param.origin == UDM_WORD_ORIGIN_QUERY)
    {
      char name[32];
      udm_snprintf(name, sizeof(name), "uword%d.word", (int) W->Param.order);
      UdmVarListReplaceStr(Vars, name, W->Word.str);
      udm_snprintf(name, sizeof(name), "uword%d.count", (int) W->Param.order);
      UdmVarListReplaceInt(Vars, name,
                           (int) UdmWWListSumCount(WWList, W->Param.order));
    }
  }
}


static void
UdmWWListWordInfoIndividualWords(UDM_VARLIST *Vars,
                                 const UDM_WIDEWORDLIST *WWList)
{
  size_t i;
  for(i= 0; i < WWList->nwords; i++)
  {
    char name[32], count[32]= "";
    UDM_WIDEWORD *W= &WWList->Word[i];
    switch (W->Param.origin)
    {
      case UDM_WORD_ORIGIN_QUERY:
      case UDM_WORD_ORIGIN_SPELL:
      case UDM_WORD_ORIGIN_SYNONYM:
      case UDM_WORD_ORIGIN_SYNONYM_FINAL:
      case UDM_WORD_ORIGIN_COLLATION:
        sprintf(count, "%d", (int) W->Param.count);
        break;
      case UDM_WORD_ORIGIN_STOP:
        strcpy(count, "stopword");
        break;
      case UDM_WORD_ORIGIN_SUGGEST:
        break;
    }
    sprintf(name, "word%d.word", (int) i);
    UdmVarListAddStr(Vars, name, W->Word.str);
    sprintf(name, "word%d.count", (int) i);
    UdmVarListAddStr(Vars, name, count);
    sprintf(name, "word%d.doccount", (int) i);
    UdmVarListAddInt(Vars, name, W->Param.doccount);
    sprintf(name, "word%d.order", (int) i);
    UdmVarListAddInt(Vars, name, W->Param.order);
    sprintf(name, "word%d.origin", (int) i);
    UdmVarListAddInt(Vars, name, W->Param.origin);
    sprintf(name, "word%d.weight", (int) i);
    UdmVarListAddInt(Vars, name, W->Param.weight);
  }
}


static void
UdmWWListWordInfoSuggestions(const UDM_WIDEWORDLIST *WWList,
                             UDM_DSTR *dstr, udm_bool_t *have_suggestions)
{
  size_t i;
  for (i= 0; i < WWList->nwords; i++)
  {
    UDM_WIDEWORD *Wi= &WWList->Word[i];
    UDM_WIDEWORD *Wb= NULL;
   
    if (Wi->Param.origin == UDM_WORD_ORIGIN_QUERY)
    {
      if (Wi->Param.count > 0)
      {
        Wb= Wi;
      }
      else
      {
        size_t j, count= 0;
        for (j= 0; j < WWList->nwords; j++)
        {
          UDM_WIDEWORD *Wj= &WWList->Word[j];
          if (Wj->Param.origin == UDM_WORD_ORIGIN_SUGGEST &&
              Wj->Param.order == Wi->Param.order && Wj->Param.count > count)
          {
            count= WWList->Word[j].Param.count;
            Wb= Wj;
            *have_suggestions= UDM_TRUE;
          }
        }
      }
    }
    else if (Wi->Param.origin == UDM_WORD_ORIGIN_STOP)
    {
      Wb= Wi;
    }
    
    if (Wb)
    {
      UdmDSTRAppendf(dstr, "%s%s", UdmDSTRLength(dstr) ? " " : "", Wb->Word.str);
    }
  }
}


static udm_rc_t
UdmWWListWordInfo(UDM_VARLIST *Vars, UDM_WIDEWORDLIST *WWList)
{
  udm_bool_t have_suggestions= UDM_FALSE;
  UDM_DSTR dstr;

  UdmVarListAddInt(Vars, "nwords", WWList->nwords);

  /* Extended format */
  UdmDSTRInit(&dstr, 128);
  UdmWWListWordInfoUniqueWords(Vars, WWList);
  UdmWWListWordInfoIndividualWords(Vars, WWList);
  UdmWWListWordInfoExtended(WWList, &dstr);  
  UdmVarListReplaceStrn(Vars, "WE", UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr));

  /* Short format */
  UdmDSTRReset(&dstr);  
  UdmWWListWordInfoShort(WWList, &dstr, UDM_TRUE);
  UdmVarListReplaceStrn(Vars, "W", UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr));

  /* Compact format */
  UdmDSTRReset(&dstr);  
  UdmWWListWordInfoShort(WWList, &dstr, UDM_FALSE);
  UdmVarListReplaceStrn(Vars, "WC", UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr));

  /* Suggestions */
  UdmDSTRReset(&dstr);
  UdmWWListWordInfoSuggestions(WWList, &dstr, &have_suggestions);
  if (have_suggestions)
    UdmVarListReplaceStrn(Vars, "WS", UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr));
  UdmDSTRFree(&dstr);

  return UDM_OK;
}


udm_rc_t
UdmQueryWordInfo(UDM_QUERY *Query)
{
  return UdmWWListWordInfo(&Query->Res.Vars, &Query->Res.WWList);
}
